/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002 CodeFactory AB
 * Copyright (C) 2001-2002 Richard Hult <richard@imendio.com>
 * Copyright (C) 2001-2002 Mikael Hallendal <micke@imendio.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

/**
 * SECTION:mrp-error
 * @Short_Description: Error reporting.
 * @Title: MrpError
 * @include: libplanner/mrp-error.h
 *
 * libplanner uses #GError to report errors.
 *
 * The #GQuark "mrp-error-quark" is used as the domain.
 * Use #MRP_ERROR macro to get it.
 *
 * #MrpError lists the error codes used.
 */

#include <config.h>
#include "mrp-error.h"

/**
 * mrp_error_quark:
 * @Returns: a #GQuark
 *
 * Get libplanner error domain.
 */
GQuark
mrp_error_quark (void)
{
	static GQuark q = 0;

	if (q == 0) {
		q = g_quark_from_static_string ("mrp-error-quark");
	}

	return q;
}
